"""
All magic keywords shall summon here
"""

MC_P_EN   = "mc_p_en"
MC_P_TY   = "mc_p_ty"
MC_P_ZE   = "mc_p_ze"
MC_P_WE   = "mc_p_we"
MC_P_GW   = "mc_p_gw"
MC_P_TS   = "mc_p_ts"
RUN_START = "run_start_mjd"
RUN_STOP  = "run_stop_mjd"
RUN       = "run"
EVENT     = "event"
DATASETS  = "dataset"
